@@ -77,6 +77,7 @@ static void on_big_icon_size_changed(FmConfig* cfg, FmFolderView* fv);
 static void on_small_icon_size_changed(FmConfig* cfg, FmFolderView* fv);
 static void on_thumbnail_size_changed(FmConfig* cfg, FmFolderView* fv);
 
+static void cancel_pending_row_activated(FmFolderView* fv);
 
 static void fm_folder_view_class_init(FmFolderViewClass *klass)
 {
@@ -206,6 +207,8 @@ FmJobErrorAction on_folder_err(FmFolder* folder, GError* err, FmJobErrorSeverity
         if( err->code == G_IO_ERROR_NOT_MOUNTED && severity < FM_JOB_ERROR_CRITICAL )
             if(fm_mount_path(parent, folder->dir_path, TRUE))
                 return FM_JOB_RETRY;
+        else if(err->code == G_IO_ERROR_FAILED_HANDLED)
+            return FM_JOB_CONTINUE;
     }
     fm_show_error(parent, err->message);
     return FM_JOB_CONTINUE;
@@ -247,9 +250,38 @@ static void on_icon_view_item_activated(ExoIconView* iv, GtkTreePath* path, FmFo
     item_clicked(fv, path, FM_FV_ACTIVATED);
 }
 
+static gboolean on_idle_tree_view_row_activated(FmFolderView* fv)
+{
+    GtkTreePath* path;
+    if(gtk_tree_row_reference_valid(fv->activated_row_ref))
+    {
+        path = gtk_tree_row_reference_get_path(fv->activated_row_ref);
+        item_clicked(fv, path, FM_FV_ACTIVATED);
+    }
+    gtk_tree_row_reference_free(fv->activated_row_ref);
+    fv->activated_row_ref = NULL;
+    fv->row_activated_idle = 0;
+    return FALSE;
+}
+
 static void on_tree_view_row_activated(GtkTreeView* tv, GtkTreePath* path, GtkTreeViewColumn* col, FmFolderView* fv)
 {
-    item_clicked(fv, path, FM_FV_ACTIVATED);
+    /* Due to GTK+ and libexo bugs, here a workaround is needed.
+     * https://bugzilla.gnome.org/show_bug.cgi?id=612802
+     * http://bugzilla.xfce.org/show_bug.cgi?id=6230
+     * Gtk+ 2.20+ changed its behavior, which is really bad.
+     * row-activated signal is now issued in the second button-press events
+     * rather than double click events. The content of the view and model
+     * gets changed in row-activated signal handler before button-press-event
+     * handling is finished, and this breaks button-press handler of ExoTreeView
+     * and causing some selection-related bugs since select function cannot be reset.*/
+
+    cancel_pending_row_activated(fv);
+    if(fv->model)
+    {
+        fv->activated_row_ref = gtk_tree_row_reference_new(GTK_TREE_MODEL(fv->model), path);
+        g_idle_add((GSourceFunc)on_idle_tree_view_row_activated, fv);
+    }
 }
 
 static void fm_folder_view_init(FmFolderView *self)
@@ -304,6 +336,8 @@ static void fm_folder_view_finalize(GObject *object)
 
     g_signal_handlers_disconnect_by_func(fm_config, on_single_click_changed, object);
 
+    cancel_pending_row_activated(self);
+
     if(self->icon_size_changed_handler)
         g_signal_handler_disconnect(fm_config, self->icon_size_changed_handler);
 
@@ -356,9 +390,11 @@ void fm_folder_view_set_mode(FmFolderView* fv, FmFolderViewMode mode)
         FmFolderModel* model = (FmFolderModel*)fv->model;
         int icon_size = 0;
         guint handler;
+        gboolean has_focus;
 
         if( G_LIKELY(fv->view) )
         {
+            has_focus = GTK_WIDGET_HAS_FOCUS(fv->view);
             /* preserve old selections */
             sels = fm_folder_view_get_selected_tree_paths(fv);
             gtk_widget_destroy(fv->view );
@@ -367,7 +403,10 @@ void fm_folder_view_set_mode(FmFolderView* fv, FmFolderViewMode mode)
              * reuse the widget when available. */
         }
         else
+        {
             sels = NULL;
+            has_focus = FALSE;
+        }
 
         if(fv->icon_size_changed_handler)
         {
@@ -528,6 +567,7 @@ void fm_folder_view_set_mode(FmFolderView* fv, FmFolderViewMode mode)
             ts = gtk_tree_view_get_selection((GtkTreeView*)fv->view);
             g_signal_connect(fv->view, "row-activated", G_CALLBACK(on_tree_view_row_activated), fv);
             g_signal_connect(ts, "changed", G_CALLBACK(on_sel_changed), fv);
+            /*cancel_pending_row_activated(fv);*/ /* FIXME: is this needed? */
             gtk_tree_view_set_model((GtkTreeView*)fv->view, fv->model);
             gtk_tree_selection_set_mode(ts, fv->sel_mode);
             for(l = sels;l;l=l->next)
@@ -553,6 +593,9 @@ void fm_folder_view_set_mode(FmFolderView* fv, FmFolderViewMode mode)
 
         gtk_widget_show(fv->view);
         gtk_container_add((GtkContainer*)fv, fv->view);
+
+        if(has_focus) /* restore the focus if needed. */
+            gtk_widget_grab_focus(fv->view);
     }
     else
 	{
@@ -650,6 +693,7 @@ static void on_folder_unmounted(FmFolder* folder, FmFolderView* fv)
     switch(fv->mode)
     {
     case FM_FV_LIST_VIEW:
+        cancel_pending_row_activated(fv);
         gtk_tree_view_set_model(GTK_TREE_VIEW(fv->view), NULL);
         break;
     case FM_FV_ICON_VIEW:
@@ -678,6 +722,7 @@ static void on_folder_loaded(FmFolder* folder, FmFolderView* fv)
     switch(fv->mode)
     {
     case FM_FV_LIST_VIEW:
+        cancel_pending_row_activated(fv);
         gtk_tree_view_set_model(GTK_TREE_VIEW(fv->view), model);
         icon_size = fm_config->small_icon_size;
         fm_folder_model_set_icon_size(model, icon_size);
@@ -744,6 +789,7 @@ gboolean fm_folder_view_chdir(FmFolderView* fv, FmPath* path)
             switch(fv->mode)
             {
             case FM_FV_LIST_VIEW:
+                cancel_pending_row_activated(fv);
                 gtk_tree_view_set_model(GTK_TREE_VIEW(fv->view), NULL);
                 break;
             case FM_FV_ICON_VIEW:
@@ -849,6 +895,7 @@ gboolean on_btn_pressed(GtkWidget* view, GdkEventButton* evt, FmFolderView* fv)
 {
     GList* sels;
     FmFolderViewClickType type = 0;
+    GtkTreePath* tp;
 
     if(!fv->model)
         return FALSE;
@@ -857,27 +904,50 @@ gboolean on_btn_pressed(GtkWidget* view, GdkEventButton* evt, FmFolderView* fv)
     if( evt->type == GDK_BUTTON_PRESS )
     {
 		/* special handling for ExoIconView */
-		if(evt->button != 1 && (fv->mode==FM_FV_ICON_VIEW || fv->mode==FM_FV_COMPACT_VIEW || fv->mode==FM_FV_THUMBNAIL_VIEW))
+		if(evt->button != 1)
 		{
-			GtkTreePath* tp;
-			/* select the item on right click for ExoIconView */
-		    if(exo_icon_view_get_item_at_pos((ExoIconView*)view, evt->x, evt->y, &tp, NULL))
-			{
-				/* if the hit item is not currently selected */
-				if(!exo_icon_view_path_is_selected((ExoIconView*)view, tp))
-				{
-					sels = exo_icon_view_get_selected_items((const struct ExoIconView *)view);
-					if( sels ) /* if there are selected items */
-					{
-						exo_icon_view_unselect_all(EXO_ICON_VIEW(view)); /* unselect all items */
-						g_list_foreach(sels, (GFunc)gtk_tree_path_free, NULL);
-						g_list_free(sels);
-					}
-					exo_icon_view_select_path((ExoIconView*)view, tp);
-					exo_icon_view_set_cursor((ExoIconView*)view, tp, NULL, FALSE);
-				}
-				gtk_tree_path_free(tp);
-			}
+            if(fv->mode==FM_FV_ICON_VIEW || fv->mode==FM_FV_COMPACT_VIEW || fv->mode==FM_FV_THUMBNAIL_VIEW)
+            {
+                /* select the item on right click for ExoIconView */
+                if(exo_icon_view_get_item_at_pos(EXO_ICON_VIEW(view), evt->x, evt->y, &tp, NULL))
+                {
+                    /* if the hit item is not currently selected */
+                    if(!exo_icon_view_path_is_selected(EXO_ICON_VIEW(view), tp))
+                    {
+                        sels = exo_icon_view_get_selected_items((const struct ExoIconView *)view);
+                        if( sels ) /* if there are selected items */
+                        {
+                            exo_icon_view_unselect_all(EXO_ICON_VIEW(view)); /* unselect all items */
+                            g_list_foreach(sels, (GFunc)gtk_tree_path_free, NULL);
+                            g_list_free(sels);
+                        }
+                        exo_icon_view_select_path(EXO_ICON_VIEW(view), tp);
+                        exo_icon_view_set_cursor(EXO_ICON_VIEW(view), tp, NULL, FALSE);
+                    }
+                    gtk_tree_path_free(tp);
+                }
+            }
+            else if( fv->mode == FM_FV_LIST_VIEW
+                     && evt->window == gtk_tree_view_get_bin_window(GTK_TREE_VIEW(view)))
+            {
+                /* special handling for ExoTreeView */
+                /* Fix #2986834: MAJOR PROBLEM: Deletes Wrong File Frequently. */
+                GtkTreeViewColumn* col;
+                if(gtk_tree_view_get_path_at_pos(GTK_TREE_VIEW(view), evt->x, evt->y, &tp, &col, NULL, NULL))
+                {
+                    GtkTreeSelection* tree_sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(view));
+                    if(!gtk_tree_selection_path_is_selected(tree_sel, tp))
+                    {
+                        gtk_tree_selection_unselect_all(tree_sel);
+                        if(col == exo_tree_view_get_activable_column(EXO_TREE_VIEW(view)))
+                        {
+                            gtk_tree_selection_select_path(tree_sel, tp);
+                            gtk_tree_view_set_cursor(GTK_TREE_VIEW(view), tp, NULL, FALSE);
+                        }
+                    }
+                    gtk_tree_path_free(tp);
+                }
+            }
 		}
 
 		if(evt->button == 2) /* middle click */
@@ -1081,3 +1151,14 @@ FmFileInfo* fm_folder_view_get_cwd_info(FmFolderView* fv)
 {
     return FM_FOLDER_MODEL(fv->model)->dir->dir_fi;
 }
+
+static void cancel_pending_row_activated(FmFolderView* fv)
+{
+    if(fv->row_activated_idle)
+    {
+        g_source_remove(fv->row_activated_idle);
+        fv->row_activated_idle = 0;
+        gtk_tree_row_reference_free(fv->activated_row_ref);
+        fv->activated_row_ref = NULL;
+    }
+}
