@@ -169,24 +169,30 @@ gchar* fm_get_user_input_rename(GtkWindow* parent, const char* title, const char
         /* only select filename part without extension name. */
         if(default_text[1])
         {
-            /* FIXME: handle the special case for *.tar.gz or *.tar.bz2 */
+            /* FIXME: handle the special case for *.tar.gz or *.tar.bz2
+             * We should exam the file extension with g_content_type_guess, and
+             * find out a longest valid extension name.
+             * For example, the extension name of foo.tar.gz is .tar.gz, not .gz. */
             const char* dot = g_utf8_strrchr(default_text, -1, '.');
-/*
-            const char* dot = default_text;
-            while( dot = g_utf8_strchr(dot + 1, -1, '.') )
-            {
-                gboolean uncertain;
-                if(g_content_type_guess(dot, NULL, 0, &uncertain))
-                {
-                    gtk_editable_select_region(entry, 0, g_utf8_pointer_to_offset(default_text, dot));
-                    break;
-                }
-            }
-*/
             if(dot)
                 gtk_editable_select_region(GTK_EDITABLE(entry), 0, g_utf8_pointer_to_offset(default_text, dot));
             else
                 gtk_editable_select_region(GTK_EDITABLE(entry), 0, -1);
+            /*
+            const char* dot = default_text;
+            while( dot = g_utf8_strchr(dot + 1, -1, '.') )
+            {
+                gboolean uncertain;
+                char* type = g_content_type_guess(dot-1, NULL, 0, &uncertain);
+                if(!g_content_type_is_unknown(type))
+                {
+                    g_free(type);
+                    gtk_editable_select_region(entry, 0, g_utf8_pointer_to_offset(default_text, dot));
+                    break;
+                }
+                g_free(type);
+            }
+            */
         }
     }
 
@@ -216,9 +222,18 @@ static GtkDialog* _fm_get_user_input_dialog(GtkWindow* parent, const char* title
 static gchar* _fm_user_input_dialog_run( GtkDialog* dlg, GtkEntry *entry)
 {
     char* str = NULL;
+    int sel_start, sel_end;
+    gboolean has_sel;
 
+    /* FIXME: this workaround is used to overcome bug of gtk+.
+     * gtk+ seems to ignore select region and select all text for entry in dialog. */
+    has_sel = gtk_editable_get_selection_bounds(GTK_EDITABLE(entry), &sel_start, &sel_end);
     gtk_box_pack_start(GTK_BOX( GTK_DIALOG(dlg)->vbox ), GTK_WIDGET( entry ), FALSE, TRUE, 6);
     gtk_widget_show_all(GTK_WIDGET(dlg));
+
+    if(has_sel)
+        gtk_editable_select_region(GTK_EDITABLE(entry), sel_start, sel_end);
+
     while(gtk_dialog_run(dlg) == GTK_RESPONSE_OK)
     {
         const char* pstr = gtk_entry_get_text(entry);
@@ -366,33 +381,40 @@ gboolean fm_do_mount(GtkWindow* parent, GObject* obj, MountAction action, gboole
     {
         if(interactive)
         {
-            if(data->err->domain == G_IO_ERROR && data->err->code == G_IO_ERROR_FAILED)
+            if(data->err->domain == G_IO_ERROR)
             {
-                /* Generate a more human-readable error message instead of using a gvfs one. */
-
-                /* The original error message is something like:
-                 * Error unmounting: umount exited with exit code 1:
-                 * helper failed with: umount: only root can unmount
-                 * UUID=18cbf00c-e65f-445a-bccc-11964bdea05d from /media/sda4 */
-
-                /* Why they pass this back to us?
-                 * This is not human-readable for the users at all. */
-
-                if(strstr(data->err->message, "only root can "))
+                if(data->err->code == G_IO_ERROR_FAILED)
                 {
-                    g_debug("%s", data->err->message);
-                    g_free(data->err->message);
-                    data->err->message = g_strdup(_("Only system administrators have the permission to do this."));
+                    /* Generate a more human-readable error message instead of using a gvfs one. */
+
+                    /* The original error message is something like:
+                     * Error unmounting: umount exited with exit code 1:
+                     * helper failed with: umount: only root can unmount
+                     * UUID=18cbf00c-e65f-445a-bccc-11964bdea05d from /media/sda4 */
+
+                    /* Why they pass this back to us?
+                     * This is not human-readable for the users at all. */
+
+                    if(strstr(data->err->message, "only root can "))
+                    {
+                        g_debug("%s", data->err->message);
+                        g_free(data->err->message);
+                        data->err->message = g_strdup(_("Only system administrators have the permission to do this."));
+                    }
                 }
+                else if(data->err->code == G_IO_ERROR_FAILED_HANDLED)
+                    interactive = FALSE;
             }
-            fm_show_error(parent, data->err->message);
+            if(interactive)
+                fm_show_error(parent, data->err->message);
         }
         g_error_free(data->err);
     }
 
     g_free(data);
     g_object_unref(cancellable);
-    g_object_unref(op);
+    if(op)
+        g_object_unref(op);
     return ret;
 }
 
@@ -442,7 +464,6 @@ gboolean fm_eject_volume(GtkWindow* parent, GVolume* vol, gboolean interactive)
 
 void fm_copy_files(FmPathList* files, FmPath* dest_dir)
 {
-	GtkWidget* dlg;
 	FmJob* job = fm_file_ops_job_new(FM_FILE_OP_COPY, files);
 	fm_file_ops_job_set_dest(FM_FILE_OPS_JOB(job), dest_dir);
     fm_file_ops_job_run_with_progress(FM_FILE_OPS_JOB(job));
@@ -450,7 +471,6 @@ void fm_copy_files(FmPathList* files, FmPath* dest_dir)
 
 void fm_move_files(FmPathList* files, FmPath* dest_dir)
 {
-	GtkWidget* dlg;
 	FmJob* job = fm_file_ops_job_new(FM_FILE_OP_MOVE, files);
 	fm_file_ops_job_set_dest(FM_FILE_OPS_JOB(job), dest_dir);
     fm_file_ops_job_run_with_progress(FM_FILE_OPS_JOB(job));
@@ -458,17 +478,21 @@ void fm_move_files(FmPathList* files, FmPath* dest_dir)
 
 void fm_trash_files(FmPathList* files)
 {
-    if(!fm_config->confirm_del || fm_yes_no(NULL, _("Do you want to move the selected files to trash bin?"), TRUE))
+    if(!fm_config->confirm_del || fm_yes_no(NULL, _("Do you want to move the selected files to trash can?"), TRUE))
     {
-    	GtkWidget* dlg;
         FmJob* job = fm_file_ops_job_new(FM_FILE_OP_TRASH, files);
         fm_file_ops_job_run_with_progress(FM_FILE_OPS_JOB(job));
     }
 }
 
+void fm_untrash_files(FmPathList* files)
+{
+    FmJob* job = fm_file_ops_job_new(FM_FILE_OP_UNTRASH, files);
+    fm_file_ops_job_run_with_progress(FM_FILE_OPS_JOB(job));
+}
+
 static void fm_delete_files_internal(FmPathList* files)
 {
-    GtkWidget* dlg;
     FmJob* job = fm_file_ops_job_new(FM_FILE_OP_DELETE, files);
     fm_file_ops_job_run_with_progress(FM_FILE_OPS_JOB(job));
 }
@@ -516,12 +540,6 @@ void fm_move_or_copy_files_to(FmPathList* files, gboolean is_move)
     }
 }
 
-/*
-void fm_rename_files(FmPathList* files)
-{
-
-}
-*/
 
 void fm_rename_file(FmPath* file)
 {
@@ -549,7 +567,7 @@ void fm_rename_file(FmPath* file)
 
 void fm_empty_trash()
 {
-    if(fm_yes_no(NULL, _("Are you sure you want to empty the trash bin?"), TRUE))
+    if(fm_yes_no(NULL, _("Are you sure you want to empty the trash can?"), TRUE))
     {
         FmPathList* paths = fm_path_list_new();
         fm_list_push_tail(paths, fm_path_get_trash());
