@@ -19,7 +19,10 @@
  *      MA 02110-1301, USA.
  */
 
+#ifdef HAVE_CONFIG_H
 #include <config.h>
+#endif
+
 #include "fm-progress-dlg.h"
 #include "fm-gtk-utils.h"
 #include <glib/gi18n-lib.h>
@@ -117,8 +120,13 @@ static void on_cur_file(FmFileOpsJob* job, const char* cur_file, FmProgressDispl
 static FmJobErrorAction on_error(FmFileOpsJob* job, GError* err, FmJobErrorSeverity severity, FmProgressDisplay* data)
 {
     GtkTextIter it;
-    if(err->domain == G_IO_ERROR && err->code == G_IO_ERROR_CANCELLED)
-        return;
+    if(err->domain == G_IO_ERROR)
+    {
+        if(err->code == G_IO_ERROR_CANCELLED)
+            return FM_JOB_ABORT;
+        else if(err->code == G_IO_ERROR_FAILED_HANDLED)
+            return FM_JOB_CONTINUE;
+    }
 
     if(data->timer)
         g_timer_stop(data->timer);
@@ -202,14 +210,14 @@ static gint on_ask_rename(FmFileOpsJob* job, FmFileInfo* src, FmFileInfo* dest,
     disp_size = fm_file_info_get_disp_size(src);
     if(disp_size)
     {
-        tmp = g_strdup_printf("Type: %s\nSize: %s\nModified: %s",
+        tmp = g_strdup_printf(_("Type: %s\nSize: %s\nModified: %s"),
                               fm_file_info_get_desc(src),
                               disp_size,
                               fm_file_info_get_disp_mtime(src));
     }
     else
     {
-        tmp = g_strdup_printf("Type: %s\nModified: %s",
+        tmp = g_strdup_printf(_("Type: %s\nModified: %s"),
                               fm_file_info_get_desc(src),
                               fm_file_info_get_disp_mtime(src));
     }
@@ -221,14 +229,14 @@ static gint on_ask_rename(FmFileOpsJob* job, FmFileInfo* src, FmFileInfo* dest,
     disp_size = fm_file_info_get_disp_size(dest);
     if(disp_size)
     {
-        tmp = g_strdup_printf("Type: %s\nSize: %s\nModified: %s",
+        tmp = g_strdup_printf(_("Type: %s\nSize: %s\nModified: %s"),
                               fm_file_info_get_desc(dest),
                               fm_file_info_get_disp_size(dest),
                               fm_file_info_get_disp_mtime(dest));
     }
     else
     {
-        tmp = g_strdup_printf("Type: %s\nModified: %s",
+        tmp = g_strdup_printf(_("Type: %s\nModified: %s"),
                               fm_file_info_get_desc(dest),
                               fm_file_info_get_disp_mtime(dest));
     }
@@ -236,7 +244,9 @@ static gint on_ask_rename(FmFileOpsJob* job, FmFileInfo* src, FmFileInfo* dest,
     gtk_label_set_text(GTK_LABEL(dest_fi), tmp);
     g_free(tmp);
 
-    gtk_entry_set_text(GTK_ENTRY(filename), dest->disp_name);
+    tmp = g_filename_display_name(dest->path->name);
+    gtk_entry_set_text(GTK_ENTRY(filename), tmp);
+    g_free(tmp);
     g_object_set_data(G_OBJECT(filename), "old_name", dest->disp_name);
     g_signal_connect(filename, "changed", on_filename_changed, gtk_builder_get_object(builder, "rename"));
 
@@ -261,7 +271,7 @@ static gint on_ask_rename(FmFileOpsJob* job, FmFileInfo* src, FmFileInfo* dest,
 
     if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(apply_all)))
     {
-        if(res == RESPONSE_OVERWRITE || res == RESPONSE_SKIP)
+        if(res == RESPONSE_OVERWRITE || res == FM_FILE_OP_SKIP)
             data->default_opt = res;
     }
 
@@ -313,6 +323,24 @@ static void on_finished(FmFileOpsJob* job, FmProgressDisplay* data)
     }
     else
         fm_progress_display_destroy(data);
+
+    /* sepcial handling for trash
+     * FIXME: need to refactor this to use a more elegant way later. */
+    if(job->type == FM_FILE_OP_TRASH)
+    {
+        FmPathList* unsupported = (FmPathList*)g_object_get_data(job, "trash-unsupported");
+        /* some files cannot be trashed because underlying filesystems don't support it. */
+        if(unsupported) /* delete them instead */
+        {
+            if(fm_yes_no(NULL, _("Some files cannot be moved to trash can because "
+                        "the underlying file systems don't support this operation.\n"
+                        "Do you want to delete them instead?"), TRUE))
+            {
+                FmJob* job = fm_file_ops_job_new(FM_FILE_OP_DELETE, unsupported);
+                fm_file_ops_job_run_with_progress(job);
+            }
+        }
+    }
 }
 
 static void on_cancelled(FmFileOpsJob* job, FmProgressDisplay* data)
