@@ -27,7 +27,7 @@
 #include <unistd.h>
 #include "fm-monitor.h"
 
-const char query[]=
+static const char query[]=
 	G_FILE_ATTRIBUTE_STANDARD_TYPE","
 	G_FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME","
 	G_FILE_ATTRIBUTE_STANDARD_NAME","
@@ -39,7 +39,7 @@ const char query[]=
 
 static void progress_cb(goffset cur, goffset total, FmFileOpsJob* job);
 
-gboolean fm_file_ops_job_copy_file(FmFileOpsJob* job, GFile* src, GFileInfo* inf, GFile* dest)
+gboolean _fm_file_ops_job_copy_file(FmFileOpsJob* job, GFile* src, GFileInfo* inf, GFile* dest)
 {
     /* FIXME: prevent copying to self or copying parent dir to child. */
     gboolean ret = FALSE;
@@ -74,10 +74,7 @@ _retry_query_src_info:
 
 	is_virtual = g_file_info_get_attribute_boolean(inf, G_FILE_ATTRIBUTE_STANDARD_IS_VIRTUAL);
     type = g_file_info_get_file_type(inf);
-/*
-    size = g_file_info_get_attribute_uint64(inf, G_FILE_ATTRIBUTE_UNIX_BLOCKS);
-    size *= g_file_info_get_attribute_uint32(inf, G_FILE_ATTRIBUTE_UNIX_BLOCK_SIZE);
-*/
+
     size = g_file_info_get_size(inf);
     mode = g_file_info_get_attribute_uint32(inf, G_FILE_ATTRIBUTE_UNIX_MODE);
 
@@ -98,12 +95,16 @@ _retry_query_src_info:
                 {
                     g_error_free(err);
                     err = NULL;
+
+                    g_object_ref(dest); /*if dest == new_dest, it will become invalid when we unref new_dest */
                     if(new_dest)
                     {
                         g_object_unref(new_dest);
                         new_dest = NULL;
                     }
                     opt = fm_file_ops_job_ask_rename(job, src, NULL, dest, &new_dest);
+                    g_object_unref(dest);
+
                     switch(opt)
                     {
                     case FM_FILE_OP_RENAME:
@@ -111,6 +112,10 @@ _retry_query_src_info:
                         goto _retry_mkdir;
                         break;
                     case FM_FILE_OP_SKIP:
+                        /* when a dir is skipped, we need to know its total size to calculate correct progress */
+                        job->finished += size;
+                        fm_file_ops_job_emit_percent(job);
+                        job->skip_dir_content = TRUE;
                         ret = FALSE;
                         break;
                     case FM_FILE_OP_OVERWRITE:
@@ -120,12 +125,6 @@ _retry_query_src_info:
                         ret = FALSE;
                         break;
                     }
-                    if( opt == FM_FILE_OP_SKIP )
-                    {
-                        job->finished += size;
-                        fm_file_ops_job_emit_percent(job);
-                        break;
-                    }
                 }
                 else if(!fm_job_is_cancelled(fmjob))
                 {
@@ -179,27 +178,36 @@ _retry_query_src_info:
                         inf = g_file_enumerator_next_file(enu, fm_job_get_cancellable(fmjob), &err);
                         if( inf )
                         {
-                            gboolean ret;
-                            GFileMonitor* old_dest_mon = job->dest_folder_mon;
-                            GFile* sub = g_file_get_child(src, g_file_info_get_name(inf));
-                            GFile* sub_dest = g_file_get_child(dest, g_file_info_get_name(inf));
-
-                            if(g_file_is_native(dest))
-                                job->dest_folder_mon = NULL;
-                            else
-                                job->dest_folder_mon = fm_monitor_lookup_dummy_monitor(dest);
-
-                            ret = fm_file_ops_job_copy_file(job, sub, inf, sub_dest);
-                            g_object_unref(sub);
-                            g_object_unref(sub_dest);
-
-                            if(job->dest_folder_mon)
-                                g_object_unref(job->dest_folder_mon);
-                            job->dest_folder_mon = old_dest_mon;
-
-                            if( G_UNLIKELY(!ret) )
+                            /* don't overwrite dir content, only calculate progress. */
+                            if(G_UNLIKELY(job->skip_dir_content))
                             {
-                                /* FIXME: error handling */
+                                job->finished += g_file_info_get_size(inf);
+                                fm_file_ops_job_emit_percent(job);
+                            }
+                            else
+                            {
+                                gboolean ret2;
+                                GFileMonitor* old_dest_mon = job->dest_folder_mon;
+                                GFile* sub = g_file_get_child(src, g_file_info_get_name(inf));
+                                GFile* sub_dest = g_file_get_child(dest, g_file_info_get_name(inf));
+
+                                if(g_file_is_native(dest))
+                                    job->dest_folder_mon = NULL;
+                                else
+                                    job->dest_folder_mon = fm_monitor_lookup_dummy_monitor(dest);
+
+                                ret2 = _fm_file_ops_job_copy_file(job, sub, inf, sub_dest);
+                                g_object_unref(sub);
+                                g_object_unref(sub_dest);
+
+                                if(job->dest_folder_mon)
+                                    g_object_unref(job->dest_folder_mon);
+                                job->dest_folder_mon = old_dest_mon;
+
+                                if( G_UNLIKELY(!ret) )
+                                {
+                                    /* FIXME: error handling */
+                                }
                             }
                             g_object_unref(inf);
                         }
@@ -232,6 +240,7 @@ _retry_query_src_info:
                         goto _retry_enum_children;
                 }
             }
+            job->skip_dir_content = FALSE;
 		}
 		break;
 
@@ -283,12 +292,15 @@ _retry_query_src_info:
             {
                 g_error_free(err);
                 err = NULL;
+
+                g_object_ref(dest); /*if dest == new_dest, it will become invalid when we unref new_dest */
                 if(new_dest)
                 {
                     g_object_unref(new_dest);
                     new_dest = NULL;
                 }
                 opt = fm_file_ops_job_ask_rename(job, src, NULL, dest, &new_dest);
+                g_object_unref(dest);
                 switch(opt)
                 {
                 case FM_FILE_OP_RENAME:
@@ -340,7 +352,7 @@ _retry_query_src_info:
 
     /* if this is a cross-device move operation, delete source files. */
     if( ret && job->type == FM_FILE_OP_MOVE )
-        ret = fm_file_ops_job_delete_file(FM_JOB(job), src, inf); /* delete the source file. */
+        ret = _fm_file_ops_job_delete_file(FM_JOB(job), src, inf); /* delete the source file. */
 
     if(new_dest)
         g_object_unref(new_dest);
@@ -348,7 +360,7 @@ _retry_query_src_info:
     return ret;
 }
 
-gboolean fm_file_ops_job_move_file(FmFileOpsJob* job, GFile* src, GFileInfo* inf, GFile* dest)
+gboolean _fm_file_ops_job_move_file(FmFileOpsJob* job, GFile* src, GFileInfo* inf, GFile* dest)
 {
     /* FIXME: prevent moving to self or moving parent dir to child. */
 	GError* err = NULL;
@@ -376,7 +388,7 @@ _retry_query_src_info:
 
     src_fs_id = g_file_info_get_attribute_string(inf, G_FILE_ATTRIBUTE_ID_FILESYSTEM);
     /* Check if source and destination are on the same device */
-    if( g_strcmp0(src_fs_id, job->dest_fs_id) == 0 ) /* same device */
+    if( job->type == FM_FILE_OP_UNTRASH || g_strcmp0(src_fs_id, job->dest_fs_id) == 0 ) /* same device */
     {
         FmFileOpOption opt = 0;
         guint64 size;
@@ -384,19 +396,21 @@ _retry_query_src_info:
 
         /* showing currently processed file. */
         fm_file_ops_job_emit_cur_file(job, g_file_info_get_display_name(inf));
-
         _retry_move:
         if( !g_file_move(src, dest, flags, fm_job_get_cancellable(fmjob), progress_cb, job, &err))
         {
             flags &= ~G_FILE_COPY_OVERWRITE;
             if(err->domain == G_IO_ERROR && err->code == G_IO_ERROR_EXISTS)
             {
+                g_object_ref(dest); /*if dest == new_dest, it will become invalid when we unref new_dest */
                 if(new_dest)
                 {
                     g_object_unref(new_dest);
                     new_dest = NULL;
                 }
                 opt = fm_file_ops_job_ask_rename(job, src, NULL, dest, &new_dest);
+                g_object_unref(dest);
+
                 g_error_free(err);
                 err = NULL;
 
@@ -408,7 +422,61 @@ _retry_query_src_info:
                     break;
                 case FM_FILE_OP_OVERWRITE:
                     flags |= G_FILE_COPY_OVERWRITE;
-                    goto _retry_move;
+                    if(g_file_info_get_file_type(inf) == G_FILE_TYPE_DIRECTORY) /* merge dirs */
+                    {
+                        GFileEnumerator* enu = g_file_enumerate_children(src, query, G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS,
+                                                                        fm_job_get_cancellable(job), &err);
+                        if(enu)
+                        {
+                            GFileInfo* child_inf;
+                            while(!fm_job_is_cancelled(job))
+                            {
+                                child_inf = g_file_enumerator_next_file(enu, fm_job_get_cancellable(job), &err);
+                                if(child_inf)
+                                {
+                                    GFile* child = g_file_get_child(src, g_file_info_get_name(child_inf));
+                                    GFile* child_dest = g_file_get_child(dest, g_file_info_get_name(child_inf));
+                                    _fm_file_ops_job_move_file(job, child, child_inf, child_dest);
+                                    g_object_unref(child);
+                                    g_object_unref(child_dest);
+                                    g_object_unref(child_inf);
+                                }
+                                else
+                                {
+                                    if(err) /* error */
+                                    {
+                                        FmJobErrorAction act = fm_job_emit_error(fmjob, err, FM_JOB_ERROR_MODERATE);
+                                        g_error_free(err);
+                                        err = NULL;
+                                    }
+                                    else /* EOF */
+                                    {
+                                        break;
+                                    }
+                                }
+                            }
+                            g_object_unref(enu);
+                        }
+                        else
+                        {
+                            FmJobErrorAction act = fm_job_emit_error(fmjob, err, FM_JOB_ERROR_MODERATE);
+                            g_error_free(err);
+                            err = NULL;
+                            /* if(act == FM_JOB_RETRY)
+                                goto _retry_move; */
+                        }
+
+                        /* remove source dir after its content is merged with destination dir */
+                        if(!g_file_delete(src, fm_job_get_cancellable(job), &err))
+                        {
+                            FmJobErrorAction act = fm_job_emit_error(fmjob, err, FM_JOB_ERROR_MODERATE);
+                            g_error_free(err);
+                            err = NULL;
+                            /* FIXME: should this be recoverable? */
+                        }
+                    }
+                    else /* the destination is a file, just overwrite it. */
+                        goto _retry_move;
                     break;
                 case FM_FILE_OP_CANCEL:
                     fm_job_cancel(FM_JOB(job));
@@ -447,8 +515,8 @@ _retry_query_src_info:
     else /* use copy if they are on different devices */
     {
         /* use copy & delete */
-        /* source file will be deleted in fm_file_ops_job_copy_file() */
-        ret = fm_file_ops_job_copy_file(job, src, inf, dest);
+        /* source file will be deleted in _fm_file_ops_job_copy_file() */
+        ret = _fm_file_ops_job_copy_file(job, src, inf, dest);
     }
 
     if(new_dest)
@@ -465,7 +533,7 @@ void progress_cb(goffset cur, goffset total, FmFileOpsJob* job)
     fm_file_ops_job_emit_percent(job);
 }
 
-gboolean fm_file_ops_job_copy_run(FmFileOpsJob* job)
+gboolean _fm_file_ops_job_copy_run(FmFileOpsJob* job)
 {
     gboolean ret = TRUE;
 	GFile *dest_dir;
@@ -505,7 +573,7 @@ gboolean fm_file_ops_job_copy_run(FmFileOpsJob* job)
 		GFile* src = fm_path_to_gfile(path);
 		GFile* dest = g_file_get_child(dest_dir, path->name);
 
-		if(!fm_file_ops_job_copy_file(job, src, NULL, dest))
+		if(!_fm_file_ops_job_copy_file(job, src, NULL, dest))
 			ret = FALSE;
         g_object_unref(src);
         g_object_unref(dest);
@@ -523,7 +591,7 @@ gboolean fm_file_ops_job_copy_run(FmFileOpsJob* job)
 	return TRUE;
 }
 
-gboolean fm_file_ops_job_move_run(FmFileOpsJob* job)
+gboolean _fm_file_ops_job_move_run(FmFileOpsJob* job)
 {
 	GFile *dest_dir;
     GFileMonitor *dest_mon;
@@ -611,7 +679,7 @@ _retry_query_dest_info:
                 job->src_folder_mon = src_mon = NULL;
         }
 
-		if(!fm_file_ops_job_move_file(job, src, NULL, dest))
+		if(!_fm_file_ops_job_move_file(job, src, NULL, dest))
 			ret = FALSE;
         g_object_unref(src);
         g_object_unref(dest);
